//used for iBTB cache replacement policy
typedef enum : UINT8
{
   WAY0 = 0,
   WAY1 = 1,
} ReplaceWay;

//Class definition for T-Raptor, can be local 
//to a thread or shared
class TRaptor
{
   public:
      //constructor
      TRaptor();
      //copy constructor 
      TRaptor(const TRaptor& r);
      //destructor
      ~TRaptor();
   
      //used for verification
      VOID PrintBCNT() const;
      VOID PrintICNT() const;
      VOID PrintAddress(const ADDRINT address) const;
      VOID PrintBHR() const;
      VOID PrintGSHAREindex(const ADDRINT address) const;
      VOID PrintGSHAREentry(const ADDRINT index) const;
      VOID PrintRASentry() const; 
      VOID PrintPIR() const;
      VOID PrintiBTBindex(const ADDRINT index) const;
      VOID PrintiBTBentry(const ADDRINT address) const;
     
      //increment counters
      inline VOID IncrementbCnt();
      inline VOID IncrementiCnt();
      //reset counters
      inline VOID ResetCounters();
 
      //get GSHARE index for check prediction
      inline ADDRINT GetGSHAREindex(const ADDRINT address) const;
      //check if correctly predicted
      inline BOOL OutcomePredictionIsCorrect(const ADDRINT index, const BOOL taken) const;
      //update GSHARE with index from GetGSHAREindex
      inline VOID UpdateGSHARE(const ADDRINT index, const BOOL taken);

      //get checks if RAS hit and updates stack pointer 
      inline BOOL RASHit(const ADDRINT target);
      //push return address on RAS
      inline VOID PushRAS(const ADDRINT address);

      //get iBTB index to see if hit
      inline ADDRINT GetIBTBindex(const ADDRINT address) const;
      //check if hit/miss
      inline BOOL IBTBisHit(const ADDRINT address, const ADDRINT index, const ADDRINT target);
      
      //local TID
      UINT8 tid;   

      //make members public
      //branch count
      UINT32 bCnt;
      //insturction count for exceptions
      UINT32 iCnt;
      //GShare
      BOOL hasGShare;
      //p is 0, 256, 512, or 1024
      UINT8 log2p;
      //Pointer to counters
      // 0 Strong NT
      // 1 Weak NT
      // 2 Weak T
      // 3 Strong T
      // if GSHARE[index] >> 1 == 1  Taken
      // if GSHARE[index] >> 1 == 0  Not Taken
      UINT8* GSHARE;
      //BHSR records outcome of last log2p, should be at least 10 bits
      UINT16 BHR;
      //used in creating GSHARE index
      ADDRINT PCmask;
      //used to mask off BHR, should be at least 10 bits
      UINT16 BHRmask;
 
      //iBTB
      BOOL hasIBTB;
      //Two way set associative
      //64 entry -> 5 bits for index
      //32 entry -> 4 bits for index
      //16 entry -> 3 bits for index
     
      //TAG is 8 bits
      //PIR is 13, 12, or 11 bits
      UINT16 PIR;
      //index can be 3, 4, or 5 bits
      UINT8 indexMask;
      //Two way set associative
      //Ways for cache
      ADDRINT* way0;
      ADDRINT* way1;
      //Replacement policy is currently LRU
      ReplaceWay  LastUsedWay;
   
      //RAS
      //RAS is 0, 8, 16, 32
      BOOL hasRAS;
      ADDRINT* RAS;
      //No more than 7  bits
      UINT8 RASp;
};



// initialize TRaptor
TRaptor::TRaptor()
{
   //asign local TID
   tid = numThreads; 
   //clear bCnt
   bCnt = 0;
   //clear iCnt
   iCnt = 0;
   // Setup GShare branch predictor
   hasGShare = true;
   //clear BHR
   BHR =  0;
   switch(GSHAREsize.Value())
   {
      case 0:
         //Doesn't use GShare
         hasGShare = false;
         break;
      case 256:
         log2p = 8;
         //mask off everything but PC[4+log2p : 4]
         PCmask = 0xFF0;
         //mask off everything but BHR[log2p : 0]
         BHRmask = 0xFF;
         break;
      case 512:
         log2p = 9;
         PCmask = 0x1FF0;
         BHRmask = 0x1FF;
         break;
      case 1024:
         log2p = 10;
         PCmask = 0x3FF0;
         BHRmask = 0x3FF;
         break;
      case 2048:
         log2p = 11;
         PCmask = 0x7FF0;
         BHRmask = 0x7FF;
         break;
      case 4096:
         log2p = 12;
         PCmask = 0xFFF0;
         BHRmask = 0xFFF;
         break;
      default:
         PIN_GetLock(&cout_lock,  PIN_ThreadId()+1);
         std::cout << "mcfTRaptor: GShare can only take on 0, 256, 512, or 1024 entries." << std::endl;
         PIN_ReleaseLock(&cout_lock);
         std::exit(-1);
   }
   if(hasGShare)
   {
      GSHARE = new UINT8[GSHAREsize.Value()];
      //Set all to weak NT
      std::fill(GSHARE, GSHARE+GSHAREsize.Value(), 0x1);
   }

   //setup RAS
   hasRAS = true;
   if(RASsize.Value() == 0)
      hasRAS = false;
   else if(RASsize.Value() == 8 || RASsize.Value() == 16 || RASsize.Value() == 32 || RASsize.Value() == 64)
   {
      RAS =  new ADDRINT[RASsize.Value()];
      RASp = 0;
   }
   else
   {
      PIN_GetLock(&cout_lock,  PIN_ThreadId()+1);
      std::cout << "mcfTRaptor: Return address stack size can only be 0, 8, 16, 32, or 64" << std::endl;
      PIN_ReleaseLock(&cout_lock);
      std::exit(-1);
   }

   //setup indirect branch target buffer
   hasIBTB = true;
   // clear path information register
   PIR = 0;
   //no iBTB
   if(IBTBsize.Value() == 0) hasIBTB = false;
   else if(IBTBsize.Value() == 16) indexMask = 0x3;
   else if(IBTBsize.Value() == 32) indexMask = 0x7; // 0b1111
   else if(IBTBsize.Value() == 64) indexMask = 0x1F; //0b1 1111
   else
   {
      PIN_GetLock(&cout_lock,  PIN_ThreadId()+1);
      std::cout << "mcfTRaptor: iBTB can only have 0, 16, 32, or 64 entries" << std::endl;
      PIN_ReleaseLock(&cout_lock);
      std::exit(-1);
   }
   if(hasIBTB)
   {
      //2-way
      UINT32 size = IBTBsize.Value()/2;
      //size is number of entries in each way
      way0 = new ADDRINT[size];
      way1 = new ADDRINT[size];
      //seed LastUsedWay
      LastUsedWay = WAY0;
   }
}

TRaptor::TRaptor(const TRaptor& r)
{
   //set members
   tid = r.tid;
   bCnt = r.bCnt;
   iCnt = r.iCnt;
   //GShare members
   hasGShare = r.hasGShare;
   log2p = r.log2p;
   BHR = r.BHR;
   PCmask = r.PCmask;
   BHRmask = r.BHRmask;
   //iBTB members
   hasIBTB = r.hasIBTB;
   PIR = r.PIR; 
   LastUsedWay = r.LastUsedWay;
   indexMask = r.indexMask;
   //RAS members
   hasRAS = r.hasRAS;
   RASp = r.RASp;

   //fill out GSHARE
   if(hasGShare)
   {
      GSHARE = new UINT8[GSHAREsize.Value()];
      std::copy(r.GSHARE, r.GSHARE+GSHAREsize.Value(), GSHARE);
   } 
   if(hasRAS)
   {
      RAS = new ADDRINT[RASsize.Value()];
      std::copy(r.RAS, r.RAS+RASsize.Value(), RAS);
   }
   if(hasIBTB)
   {
      UINT32 size = IBTBsize.Value()/2;
      way0 = new ADDRINT[size];
      way1 = new ADDRINT[size];
      std::copy(r.way0, r.way0+size, way0);
      std::copy(r.way1, r.way1+size, way1);
   }
}

TRaptor::~TRaptor()
{
   //delete allocated members
   if( hasGShare )
      delete[] GSHARE;
   if( hasRAS )
      delete[] RAS;
   if( hasIBTB )
   {
      delete[] way0;
      delete[] way1;
   }
}
VOID TRaptor::PrintAddress(const ADDRINT address) const
{
   std::cout << "Instruction address: " << std::hex << std::setw(ADDRWID) << std::setfill('0') << address << std::endl;
}
VOID TRaptor::PrintBCNT() const
{
   std::cout << "bCnt: " << bCnt << std::endl;
}
VOID TRaptor::PrintICNT() const
{
   std::cout << "iCnt: " << iCnt << std::endl;
}
VOID TRaptor::PrintBHR() const
{
   std::bitset<16> hist(BHR);
   std::cout << "BHR: 0b" << hist; 
   std::cout << " (0x" << std::hex << std::setw(4) << std::setfill('0') << BHR << std::endl;
}
VOID TRaptor::PrintGSHAREindex(const ADDRINT address) const
{ 
   std::cout << "index = ( ";
   std::cout << "0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << (PCmask>>4) << " & ";
   std::cout << "0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << (address>>4) << " ) ^ (";
   std::cout << "0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << BHR << " & ";
   std::cout << "0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << BHRmask << " ) "; 
   std::cout << std::endl;
   ADDRINT temp = ( (this->PCmask & address) >> 4 ) ^ (this->BHR & this->BHRmask);
   std::cout << "index = " << temp << std::endl;
}
VOID TRaptor::PrintGSHAREentry(const ADDRINT index) const
{
   std::cout << "GSHARE[" << index << "]: " << static_cast<int>(GSHARE[index]);
   std::string type;
   if(GSHARE[index] ==  0) type = " (SNT)";
   else if(GSHARE[index] == 1) type = " (WNT)";
   else if(GSHARE[index] == 2) type = " (WT)";
   else if(GSHARE[index] == 3) type = " (ST)";
   else type = " (Invalid value)";
   
   std::cout << type << std::endl;
}
VOID TRaptor::PrintRASentry() const
{
    std::cout << "RAS[" << RASp << "] = " << RAS[RASp] << std::endl;
}
VOID TRaptor::PrintPIR() const
{

}
VOID TRaptor::PrintiBTBindex(const ADDRINT address) const
{

}
VOID TRaptor::PrintiBTBentry(const ADDRINT index) const
{


}

//increment counters
inline VOID TRaptor::IncrementbCnt()
{
   this->bCnt++;   
}
inline VOID TRaptor::IncrementiCnt()
{
   this->iCnt++;
}
//reset counters
inline VOID TRaptor::ResetCounters()
{
   this->bCnt = 0;
   this->iCnt = 0;
}
//GSHARE member functions
//get GSHARE index
inline ADDRINT TRaptor::GetGSHAREindex(const ADDRINT address) const
{
   //create GShare index
   //GShare.index = BHR[log2(p):0] xor PC[4+log2(p):4]
   if( hasGShare )
      return ( (this->PCmask & address) >> 4 ) ^ (this->BHR & this->BHRmask);
   return 0;
}

//Check if prediction is correct
inline BOOL TRaptor::OutcomePredictionIsCorrect(const ADDRINT index, const BOOL taken) const
{
   //return false if TRaptor isn't using GSHARE
   if( !this->hasGShare )
      return false;

   //get prediction from GSHARE table
   //pred == 1 -> taken
   UINT64 pred = this->GSHARE[index] >> 1;
   if( taken == pred )
      return true;
   else 
      return false;   
}
inline void TRaptor::UpdateGSHARE(const ADDRINT index, const BOOL taken)
{
   //if not using GShare return
   if( !this->hasGShare )
      return;

   //update BHR
   this->BHR = (this->BHR << 1) | taken;
   //mask off unecessary BHR bits
   //log2p 8,9, or 10
   UINT16 temp = (1 << log2p) - 1;
   this->BHR &= temp; 

   //update gshare
   UINT8 prev = this->GSHARE[index];
   UINT8 B1 =  prev >> 1;
   UINT8 B0 =  prev & 0x1;
   //reduced
   UINT8 F1 = ((B1 & B0) | (taken & B1) | (taken &B0)) << 1;
   UINT8 F0 = (B1 & ~B0) | (taken & ~B0) | (taken & B1);
   this->GSHARE[index] = F1 | F0;
}
//Get top of RAS
inline BOOL TRaptor::RASHit(const ADDRINT target)
{
   if( !this->hasRAS )
      return false;

   ADDRINT ret = RAS[RASp];
   //update pointer
   if(RASp > 0)
      RASp--;
   
   return (target == ret);
}
inline VOID TRaptor::PushRAS(const ADDRINT address)
{

   if( !this->hasRAS )
      return;

   //if at bottom, return to top
   RASp = ( RASp == RASsize.Value() - 1 ) ? 0 : RASp + 1;
   RAS[RASp] = address;
}
//get iBTB Index
inline ADDRINT TRaptor::GetIBTBindex(const ADDRINT address) const
{
   // iBTB.index = PIR[8+index size-1: 8] xor PC[4+index size-1:4];
   if( this->hasIBTB )
      return ( ( PIR >> 8 ) & indexMask ) ^ ( ( address >> 4 ) & indexMask );
   else
      return 0;
}
inline BOOL TRaptor::IBTBisHit(const ADDRINT address, const ADDRINT index, const ADDRINT target) 
{
   //return fals if not using IBTB
   if( !this->hasIBTB )
      return false;
   
   //if hit
   BOOL hit;
   if( this->way0[index] == target )
   { 
      this->LastUsedWay = WAY0;
      hit = true;
   }
   else if( way1[index] == target )
   {
      this->LastUsedWay = WAY1;
      hit = true;
   }
   //evict
   else 
   { 
      hit = false;
      //replace LRU
      if( this->LastUsedWay == WAY0 )
      {
         this->way1[index] = target;
         this->LastUsedWay = WAY1;
      }
      else
      {
         this->way0[index] = target;
         this->LastUsedWay = WAY0;
      }
   }

   //update PIR
   //PIR[12:0] = ((PIR[12:0]<<2) xor PC[16:4]) | Outcome
   this->PIR = (  (this->PIR << 2) ^  ( ( address >> 4 ) & 0xff ) ) | 1;
   return hit;
}

