/*BEGIN_LEGAL 
Intel Open Source License 

Copyright (c) 2002-2005 Intel Corporation 
All rights reserved. 
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.  Redistributions
in binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.  Neither the name of
the Intel Corporation nor the names of its contributors may be used to
endorse or promote products derived from this software without
specific prior written permission.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE INTEL OR
ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
END_LEGAL */

//mcfTRaptorPrivateAnalysis.h
//Pin analysis routines using thread local T-Raptor
//
//Bugs and criticisms can be directed to
//myers.albert@gmail.com


//Get private T-Raptor structure
TRaptor* get_tls(THREADID threadid)
{
   TRaptor * tdata = static_cast<TRaptor*>(PIN_GetThreadData(tls_key, threadid));
   return tdata;
}

/*------------------------*/
/* ASCII Analysis routines */
/*------------------------*/
VOID Private_DirectConditional_ASCII(const THREADID threadid, const ADDRINT addr, const BOOL taken)
{
   //see if we can emit
   if(!CanEmit(threadid)) return;

   //get private TRaptor structure
   TRaptor* traptor = get_tls(threadid);
   
   //increment bCnt and iCnt
   traptor->IncrementiCnt();
   traptor->IncrementbCnt();

   ADDRINT index = traptor->GetGSHAREindex(addr);

   //see if prediction is correct
   BOOL correct = traptor->OutcomePredictionIsCorrect(index, taken);

   //update statistics for DirCB
   IncrementBranchStatistics(Outcome, correct);

   //Emit descriptor if prediction is wrong or doesn't use GShare
   if( !correct )
   {
      std::ostringstream convert;
      convert << static_cast<UINT16>(traptor->tid)  << ", " << traptor->bCnt << std::endl;
      std::string mispredicted_outcome = convert.str();
      PushASCIIDescriptor(threadid, mispredicted_outcome);
      //reset iCnt, bCnt
      traptor->ResetCounters();
   }

   //update TRaptor's GSHARE
   traptor->UpdateGSHARE(index, taken);
}
//return instruction, need to check RAS
VOID Private_Ret_ASCII(const THREADID threadid, const ADDRINT addr, const ADDRINT target)
{
   //see if we can emit
   if( !CanEmit(threadid) ) return;

   //get private TRaptor structure
   TRaptor* traptor = get_tls(threadid);   
   //increment bCnt and iCnt
   traptor->IncrementiCnt();
   traptor->IncrementbCnt();

   //check if RAS hits or doesn't use RAS
   BOOL correct = traptor->RASHit(target);

   //increment statistics for IndUB
   IncrementBranchStatistics(Target, correct);

   if( !correct )
   {

      std::ostringstream convert;
      convert << static_cast<UINT16>(traptor->tid) << ", " << traptor->bCnt;
      convert << ", T, 0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << target << std::endl;
      std::string mispredicted_target = convert.str();
      PushASCIIDescriptor(threadid, mispredicted_target);
      //reset iCnt, bCnt
      traptor->ResetCounters();
   }
}
VOID Private_IndirectCall_ASCII(const THREADID threadid, const ADDRINT addr, const ADDRINT target, const ADDRINT ret)
{
   //see if we can emit
   if(!CanEmit(threadid)) return;

   //get private TRaptor structure
   TRaptor* traptor = get_tls(threadid);   
   //increment bCnt and iCnt
   traptor->IncrementiCnt();
   traptor->IncrementbCnt();

   ADDRINT index = traptor->GetIBTBindex(addr);
 
   //check if iBTB is hit or if not using iBTB
   BOOL correct = traptor->IBTBisHit(addr, index, target);  
   //increment statistics for IndUB
   IncrementBranchStatistics(Target, correct);

   if ( !correct )
   {
      //write ASCII string
      std::ostringstream convert;
      convert << static_cast<UINT16>(traptor->tid) << ", " << traptor->bCnt;
      convert << ", T, 0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << target << std::endl;
      std::string mispredicted_target = convert.str();
      PushASCIIDescriptor(threadid, mispredicted_target);
      //reset iCnt, bCnt
      traptor->ResetCounters();
   }
   
   traptor->PushRAS(ret);
}
//other indirect branch
VOID Private_OtherUnconditionalIndirect_ASCII(const THREADID threadid, const ADDRINT addr, const ADDRINT target, const ADDRINT ret)
{
   //see if we can emit
   if(!CanEmit(threadid)) return;

   //get private TRaptor structure
   TRaptor* traptor = get_tls(threadid);   
   //increment bCnt and iCnt
   traptor->IncrementiCnt();
   traptor->IncrementbCnt();

   //get iBTB index 
   ADDRINT index = traptor->GetIBTBindex(addr);

   //check if iBTB is hit or if not using iBTB
   BOOL correct = traptor->IBTBisHit(addr, index, target);  
   //increment statistics for IndUB
   IncrementBranchStatistics(Target, correct);

   if ( !correct )
   {
      //write ASCII string
      std::ostringstream convert;
      convert << static_cast<UINT16>(traptor->tid) << ", " << traptor->bCnt;
      convert << ", T, 0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << target << std::endl;
      std::string mispredicted_target = convert.str();
      PushASCIIDescriptor(threadid, mispredicted_target);
      //reset iCnt, bCnt
      traptor->ResetCounters();
   }
}

/*------------------------*/
/* Dis Analysis routines */
/*------------------------*/
VOID Private_DirectConditional_Dis(const THREADID threadid, const ADDRINT addr, const BOOL taken)
{
   //see if we can emit
   if(!CanEmit(threadid)) return;

   //get private TRaptor structure
   TRaptor* traptor = get_tls(threadid);
   
   //increment bCnt and iCnt
   traptor->IncrementiCnt();
   traptor->IncrementbCnt();

   ADDRINT index = traptor->GetGSHAREindex(addr);

   //see if prediction is correct
   BOOL correct = traptor->OutcomePredictionIsCorrect(index, taken);

   //update statistics for DirCB
   IncrementBranchStatistics(Outcome, correct);

   //Emit descriptor if prediction is wrong or doesn't use GShare
   if( !correct )
   {
      std::string ins;
      getAssemblyString(ins, threadid, addr);

      std::ostringstream convert;
      convert << static_cast<UINT16>(traptor->tid)  << ", " << traptor->bCnt;
      convert << "\t" << ins << std::endl;
      std::string mispredicted_outcome = convert.str();
      PushASCIIDescriptor(threadid, mispredicted_outcome);
      //reset iCnt, bCnt
      traptor->ResetCounters();
   }

   //update TRaptor's GSHARE
   traptor->UpdateGSHARE(index, taken);
}
//return instruction, need to check RAS
VOID Private_Ret_Dis(const THREADID threadid, const ADDRINT addr, const ADDRINT target)
{
   //see if we can emit
   if( !CanEmit(threadid) ) return;

   //get private TRaptor structure
   TRaptor* traptor = get_tls(threadid);   
   //increment bCnt and iCnt
   traptor->IncrementiCnt();
   traptor->IncrementbCnt();

   //check if RAS hits or doesn't use RAS
   BOOL correct = traptor->RASHit(target);

   //increment statistics for IndUB
   IncrementBranchStatistics(Target, correct);

   if( !correct )
   {
      std::string ins;
      getAssemblyString(ins, threadid, addr);

      std::ostringstream convert;
      convert << static_cast<UINT16>(traptor->tid) << ", " << traptor->bCnt;
      convert << ", T, 0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << target;
      convert << "\t" << ins << std::endl;
      std::string mispredicted_target = convert.str();
      PushASCIIDescriptor(threadid, mispredicted_target);
      //reset iCnt, bCnt
      traptor->ResetCounters();
   }
}
VOID Private_IndirectCall_Dis(const THREADID threadid, const ADDRINT addr, const ADDRINT target, const ADDRINT ret)
{
   //see if we can emit
   if(!CanEmit(threadid)) return;

   //get private TRaptor structure
   TRaptor* traptor = get_tls(threadid);   
   //increment bCnt and iCnt
   traptor->IncrementiCnt();
   traptor->IncrementbCnt();

   ADDRINT index = traptor->GetIBTBindex(addr);
 
   //check if iBTB is hit or if not using iBTB
   BOOL correct = traptor->IBTBisHit(addr, index, target);  
   //increment statistics for IndUB
   IncrementBranchStatistics(Target, correct);

   if ( !correct )
   {
      std::string ins;
      getAssemblyString(ins, threadid, addr);

      //write ASCII string
      std::ostringstream convert;
      convert << static_cast<UINT16>(traptor->tid) << ", " << traptor->bCnt;
      convert << ", T, 0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << target;
      convert << "\t" << ins << std::endl;
      std::string mispredicted_target = convert.str();
      PushASCIIDescriptor(threadid, mispredicted_target);
      //reset iCnt, bCnt
      traptor->ResetCounters();
   }
   
   traptor->PushRAS(ret);
}
//other indirect branch
VOID Private_OtherUnconditionalIndirect_Dis(const THREADID threadid, const ADDRINT addr, const ADDRINT target, const ADDRINT ret)
{
   //see if we can emit
   if(!CanEmit(threadid)) return;

   //get private TRaptor structure
   TRaptor* traptor = get_tls(threadid);   
   //increment bCnt and iCnt
   traptor->IncrementiCnt();
   traptor->IncrementbCnt();

   //get iBTB index 
   ADDRINT index = traptor->GetIBTBindex(addr);

   //check if iBTB is hit or if not using iBTB
   BOOL correct = traptor->IBTBisHit(addr, index, target);  
   //increment statistics for IndUB
   IncrementBranchStatistics(Target, correct);

   if ( !correct )
   {
      std::string ins;
      getAssemblyString(ins, threadid, addr);

      //write ASCII string
      std::ostringstream convert;
      convert << static_cast<UINT16>(traptor->tid) << ", " << traptor->bCnt;
      convert << ", T, 0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << target;
      convert << "\t" << ins << std::endl; 
      std::string mispredicted_target = convert.str();
      PushASCIIDescriptor(threadid, mispredicted_target);
      //reset iCnt, bCnt
      traptor->ResetCounters();
   }
}
/*------------------------*/
/* Bin Analysis routines */
/*------------------------*/
VOID Private_DirectConditional_Bin(const THREADID threadid, const ADDRINT addr, const BOOL taken)
{
   //see if we can emit
   if(!CanEmit(threadid)) return;

   //get private TRaptor structure
   TRaptor* traptor = get_tls(threadid);
   
   //increment bCnt and iCnt
   traptor->IncrementiCnt();
   traptor->IncrementbCnt();

   ADDRINT index = traptor->GetGSHAREindex(addr);

   //see if prediction is correct
   BOOL correct = traptor->OutcomePredictionIsCorrect(index, taken);

   //update statistics for DirCB
   IncrementBranchStatistics(Outcome, correct);

   //Emit descriptor if prediction is wrong or doesn't use GShare
   if( !correct )
   {
      BinaryDescriptorTableEntry mispredicted_outcome;
      mispredicted_outcome.tid = traptor->tid;
      mispredicted_outcome.bCnt = traptor->bCnt;
      mispredicted_outcome.type = Outcome;
      PushBinDescriptor(threadid, mispredicted_outcome);
      //reset iCnt, bCnt
      traptor->ResetCounters();
   }
   //update TRaptor's GSHARE
   traptor->UpdateGSHARE(index, taken);
}
//return instruction, need to check RAS
VOID Private_Ret_Bin(const THREADID threadid, const ADDRINT addr, const ADDRINT target)
{
   //see if we can emit
   if( !CanEmit(threadid) ) return;

   //get private TRaptor structure
   TRaptor* traptor = get_tls(threadid);   
   //increment bCnt and iCnt
   traptor->IncrementiCnt();
   traptor->IncrementbCnt();

   //check if RAS hits or doesn't use RAS
   BOOL correct = traptor->RASHit(target);

   //increment statistics for IndUB
   IncrementBranchStatistics(Target, correct);

   if( !correct )
   {
      BinaryDescriptorTableEntry mispredicted_outcome;
      mispredicted_outcome.tid = traptor->tid;
      mispredicted_outcome.bCnt = traptor->bCnt;
      mispredicted_outcome.targetAddress = target;
      mispredicted_outcome.type = Target;
      PushBinDescriptor(threadid, mispredicted_outcome);
      //reset iCnt, bCnt
      traptor->ResetCounters();
   }
}
VOID Private_IndirectCall_Bin(const THREADID threadid, const ADDRINT addr, const ADDRINT target, const ADDRINT ret)
{
   //see if we can emit
   if(!CanEmit(threadid)) return;

   //get private TRaptor structure
   TRaptor* traptor = get_tls(threadid);   
   //increment bCnt and iCnt
   traptor->IncrementiCnt();
   traptor->IncrementbCnt();

   ADDRINT index = traptor->GetIBTBindex(addr);
 
   //check if iBTB is hit or if not using iBTB
   BOOL correct = traptor->IBTBisHit(addr, index, target);  
   //increment statistics for IndUB
   IncrementBranchStatistics(Target, correct);

   if ( !correct )
   {
      BinaryDescriptorTableEntry mispredicted_outcome;
      mispredicted_outcome.tid = traptor->tid;
      mispredicted_outcome.bCnt = traptor->bCnt;
      mispredicted_outcome.targetAddress = target;
      mispredicted_outcome.type = Target;
      PushBinDescriptor(threadid, mispredicted_outcome);
      //reset iCnt, bCnt
      traptor->ResetCounters();
   }
   
   traptor->PushRAS(ret);
}
//other indirect branch
VOID Private_OtherUnconditionalIndirect_Bin(const THREADID threadid, const ADDRINT addr, const ADDRINT target, const ADDRINT ret)
{
   //see if we can emit
   if(!CanEmit(threadid)) return;

   //get private TRaptor structure
   TRaptor* traptor = get_tls(threadid);   
   //increment bCnt and iCnt
   traptor->IncrementiCnt();
   traptor->IncrementbCnt();

   //get iBTB index 
   ADDRINT index = traptor->GetIBTBindex(addr);

   //check if iBTB is hit or if not using iBTB
   BOOL correct = traptor->IBTBisHit(addr, index, target);  
   //increment statistics for IndUB
   IncrementBranchStatistics(Target, correct);

   if ( !correct )
   {
      BinaryDescriptorTableEntry mispredicted_outcome;
      mispredicted_outcome.tid = traptor->tid;
      mispredicted_outcome.bCnt = traptor->bCnt;
      mispredicted_outcome.targetAddress = target;
      mispredicted_outcome.type = Target;
      PushBinDescriptor(threadid, mispredicted_outcome);
      //reset iCnt, bCnt
      traptor->ResetCounters();
   }
}

/*--------------------------------------*/
/* Other Analaysis Calls */
/*--------------------------------------*/
VOID Private_Sig(THREADID threadid, CONTEXT_CHANGE_REASON reason, const CONTEXT *ctxtFrom, CONTEXT *ctxTo, INT32 sig, VOID *v)
{
   if( !CanEmit(threadid) ) return;

   //From the Pin user guide
   //1. If the tool acquires any locks from within a Pin call-back, it must release those locks before returning from that call-back.
   //Holding a lock across Pin call-backs violates the hierarchy with respect to the Pin internal locks.
   //
   //2. If the tool calls a Pin API from within a Pin call-back or analysis routine, it should not hold any tool locks when calling the
   //API. Some of the Pin APIs use the internal Pin locks so holding a tool lock before invoking these APIs violates the hierarchy with
   //respect to the Pin internal locks.

   TRaptor* traptor = get_tls(threadid);
   traptor->IncrementiCnt();  
 
   PIN_GetLock(&cout_lock, threadid+1);
   std::cout << "mcfTRaptor: Thread " << static_cast<UINT16>(traptor->tid) << " received signal " << sig << std::endl;
   if(reason == CONTEXT_CHANGE_REASON_FATALSIGNAL)
   {
      std::cout << "mcfTRaptor: Fatal signal, exiting immediately..." << std::endl;
      PIN_GetLock(&table_lock, threadid+1);
      closeOutput();
      PIN_ReleaseLock(&table_lock);
      PIN_ReleaseLock(&cout_lock);
      PIN_ReleaseLock(&shared_lock);
      return;
   }
   PIN_ReleaseLock(&cout_lock);

   //get context
   ADDRINT from = PIN_GetContextReg(ctxtFrom, REG_INST_PTR);
   ADDRINT to = PIN_GetContextReg(ctxTo, REG_INST_PTR);

   if( ASCII.Value() )
   {
      std::ostringstream convert;
      convert << static_cast<UINT16>(traptor->tid);
      convert << ", 0";
      convert << ", " << traptor->iCnt;
      convert << ", " << std::hex << std::setw(ADDRWID) << std::setfill('0') << to;
   
      if( AnnotateDisassembly.Value() )
      {
         std::string ins;
         getAssemblyString(ins, threadid, from);
 
         if(reason==CONTEXT_CHANGE_REASON_SIGNAL)
            ins = "\tSignal At: " + ins;
         else if(reason==CONTEXT_CHANGE_REASON_SIGRETURN)
            ins = "\tReturning from signal handler: " + ins;
         convert << ins;
      }
      convert << std::endl;  
      std::string except = convert.str();
      PushASCIIDescriptor(threadid, except);
   }
   else
   {
      BinaryDescriptorTableEntry binEntry;
      binEntry.type = Exception;
      binEntry.tid = traptor->tid;
      binEntry.iCnt = traptor->iCnt;
      binEntry.targetAddress = to;
      PushBinDescriptor(threadid, binEntry);
   } 
   traptor->ResetCounters();
}

//for direct calls we still need to push target on the RAS
VOID Private_Direct_Call(THREADID threadid, ADDRINT addr, ADDRINT ret)
{
   if(!CanEmit(threadid)) return;
   TRaptor* traptor = get_tls(threadid);
   traptor->IncrementiCnt();
   traptor->PushRAS(ret);
}
//increment iCnt for all other other instructions
VOID Private_iCnt_Increment(THREADID threadid)
{
   if(!CanEmit(threadid)) return;
   TRaptor* traptor = get_tls(threadid);
   traptor->IncrementiCnt();
}

//used to decrement fastfoward_count and trace length for all instructions
VOID SetFastForwardAndLength(const THREADID threadid)
{
   //critical section
   PIN_GetLock(&count_lock, threadid+1);
   //if file limit is reacher or trace limit is reached
   //exit tool
   if( (traceLength + 1) == Length.Value() && !noLength )
   {
      PIN_Detach();
      PIN_ReleaseLock(&count_lock);
      return;
   }
   //decrement fast forward count   
   if( fastforwardLength < Skip.Value())
      fastforwardLength++;

   //otherwise increment length
   else
      traceLength++;

   PIN_ReleaseLock(&count_lock);
   return;
}


