/*BEGIN_LEGAL 
Intel Open Source License 

Copyright (c) 2002-2005 Intel Corporation 
All rights reserved. 
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.  Redistributions
in binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.  Neither the name of
the Intel Corporation nor the names of its contributors may be used to
endorse or promote products derived from this software without
specific prior written permission.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE INTEL OR
ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
END_LEGAL */

//mcfTRaptorSharedAnalysis.h
//Pin analysis routines using thread local T-Raptor
//
//Bugs and criticisms can be directed to
//myers.albert@gmail.com

//Shared T-Raptor object
TRaptor* SharedTRaptor;

//still need to use private bCnt and iCnt
class localCounters
{
   public:
      UINT8 tid;
      UINT32 bCnt;
      UINT32 iCnt; 
      localCounters(UINT32 id){ tid = id; iCnt = 0; bCnt = 0; }
      void ResetCounters(){ iCnt = 0; bCnt = 0; }
      void IncrementiCnt() { iCnt++; }
      void IncrementbCnt() { bCnt++; }
};

/*------------------------*/
/* ASCII Analysis routines */
/*------------------------*/
VOID Shared_DirectConditional_ASCII(const THREADID threadid, const ADDRINT addr, const BOOL taken)
{
   //see if we can emit
   if(!CanEmit(threadid)) return;

   localCounters *counters = static_cast<localCounters*>(PIN_GetThreadData(tls_key, threadid));
   PIN_GetLock(&shared_lock, threadid+1);
   
   //increment bCnt and iCnt
   counters->IncrementiCnt();
   counters->IncrementbCnt();

   ADDRINT index = SharedTRaptor->GetGSHAREindex(addr);

   //see if prediction is correct
   BOOL correct = SharedTRaptor->OutcomePredictionIsCorrect(index, taken);

   //update statistics for DirCB
   IncrementBranchStatistics(Outcome, correct);

   //Emit descriptor if prediction is wrong or doesn't use GShare
   if( !correct )
   {
      std::ostringstream convert;
      convert << static_cast<UINT16>(counters->tid) << ", " << counters->bCnt << std::endl;
      std::string mispredicted_outcome = convert.str();
      PushASCIIDescriptor(threadid, mispredicted_outcome);
      //reset iCnt, bCnt
      counters->ResetCounters();
   }

   //update TRaptor's GSHARE
   SharedTRaptor->UpdateGSHARE(index, taken);
   PIN_ReleaseLock(&shared_lock);
}

//return instruction, need to check RAS
VOID Shared_Ret_ASCII(const THREADID threadid, const ADDRINT addr, const ADDRINT target)
{
   //see if we can emit
   if( !CanEmit(threadid) ) return;

   localCounters *counters = static_cast<localCounters*>(PIN_GetThreadData(tls_key, threadid));
   PIN_GetLock(&shared_lock, threadid+1);
   //increment bCnt and iCnt
   counters->IncrementiCnt();
   counters->IncrementbCnt();

   //check if RAS hits or doesn't use RAS
   BOOL correct = SharedTRaptor->RASHit(target);

   //increment statistics for IndUB
   IncrementBranchStatistics(Target, correct);

   if( !correct )
   {
      std::ostringstream convert;
      convert << static_cast<UINT16>(counters->tid) << ", " << counters->bCnt;
      convert << ", T, 0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << target << std::endl;
      std::string mispredicted_target = convert.str();
      PushASCIIDescriptor(threadid, mispredicted_target);
      //reset iCnt, bCnt
      counters->ResetCounters();
   }
   PIN_ReleaseLock(&shared_lock);
}

VOID Shared_IndirectCall_ASCII(const THREADID threadid, const ADDRINT addr, const ADDRINT target, const ADDRINT ret)
{
   //see if we can emit
   if(!CanEmit(threadid)) return;

   localCounters *counters = static_cast<localCounters*>(PIN_GetThreadData(tls_key, threadid));
   PIN_GetLock(&shared_lock, threadid+1);

   //increment bCnt and iCnt
   counters->IncrementiCnt();
   counters->IncrementbCnt();

   ADDRINT index = SharedTRaptor->GetIBTBindex(addr);
 
   //check if iBTB is hit or if not using iBTB
   BOOL correct = SharedTRaptor->IBTBisHit(addr, index, target);  
   //increment statistics for IndUB
   IncrementBranchStatistics(Target, correct);

   if ( !correct )
   {
      //write ASCII string
      std::ostringstream convert;
      convert << static_cast<UINT16>(counters->tid) << ", " << counters->bCnt;
      convert << ", T, 0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << target << std::endl;
      std::string mispredicted_target = convert.str();
      PushASCIIDescriptor(threadid, mispredicted_target);
      //reset iCnt, bCnt
      counters->ResetCounters();
   }
   
   SharedTRaptor->PushRAS(ret);
   PIN_ReleaseLock(&shared_lock);
}
//other indirect branch
VOID Shared_OtherUnconditionalIndirect_ASCII(const THREADID threadid, const ADDRINT addr, const ADDRINT target, const ADDRINT ret)
{
   //see if we can emit
   if(!CanEmit(threadid)) return;

   localCounters *counters = static_cast<localCounters*>(PIN_GetThreadData(tls_key, threadid));
   PIN_GetLock(&shared_lock, threadid+1);

   //increment bCnt and iCnt
   counters->IncrementiCnt();
   counters->IncrementbCnt();

   //get iBTB index 
   ADDRINT index = SharedTRaptor->GetIBTBindex(addr);

   //check if iBTB is hit or if not using iBTB
   BOOL correct = SharedTRaptor->IBTBisHit(addr, index, target);  
   //increment statistics for IndUB
   IncrementBranchStatistics(Target, correct);

   if ( !correct )
   {
      //write ASCII string
      std::ostringstream convert;
      convert << static_cast<UINT16>(counters->tid) << ", " << counters->bCnt;
      convert << ", T, 0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << target << std::endl;
      std::string mispredicted_target = convert.str();
      PushASCIIDescriptor(threadid, mispredicted_target);
      //reset iCnt, bCnt
      counters->ResetCounters();
   }
   PIN_ReleaseLock(&shared_lock);
}

/*------------------------*/
/* Dis Analysis routines */
/*------------------------*/
VOID Shared_DirectConditional_Dis(const THREADID threadid, const ADDRINT addr, const BOOL taken)
{
   //see if we can emit
   if(!CanEmit(threadid)) return;
   localCounters *counters = static_cast<localCounters*>(PIN_GetThreadData(tls_key, threadid));
   PIN_GetLock(&shared_lock, threadid+1);

   
   //increment bCnt and iCnt
   counters->IncrementiCnt();
   counters->IncrementbCnt();

   ADDRINT index = SharedTRaptor->GetGSHAREindex(addr);

   //see if prediction is correct
   BOOL correct = SharedTRaptor->OutcomePredictionIsCorrect(index, taken);

   //update statistics for DirCB
   IncrementBranchStatistics(Outcome, correct);

   //Emit descriptor if prediction is wrong or doesn't use GShare
   if( !correct )
   {
      std::string ins;
      getAssemblyString(ins, threadid, addr);

      std::ostringstream convert;
      convert << static_cast<UINT16>(counters->tid)  << ", " << counters->bCnt;
      convert << "\t" << ins << std::endl;
      std::string mispredicted_outcome = convert.str();
      PushASCIIDescriptor(threadid, mispredicted_outcome);
      //reset iCnt, bCnt
      counters->ResetCounters();
   }

   //update TRaptor's GSHARE
   SharedTRaptor->UpdateGSHARE(index, taken);
   PIN_ReleaseLock(&shared_lock);
}
//return instruction, need to check RAS
VOID Shared_Ret_Dis(const THREADID threadid, const ADDRINT addr, const ADDRINT target)
{
   //see if we can emit
   if( !CanEmit(threadid) ) return;
   localCounters *counters = static_cast<localCounters*>(PIN_GetThreadData(tls_key, threadid));
   PIN_GetLock(&shared_lock, threadid+1);

   //increment bCnt and iCnt
   counters->IncrementiCnt();
   counters->IncrementbCnt();

   //check if RAS hits or doesn't use RAS
   BOOL correct = SharedTRaptor->RASHit(target);

   //increment statistics for IndUB
   IncrementBranchStatistics(Target, correct);

   if( !correct )
   {
      std::string ins;
      getAssemblyString(ins, threadid, addr);

      std::ostringstream convert;
      convert << static_cast<UINT16>(counters->tid) << ", " << counters->bCnt;
      convert << ", T, 0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << target;
      convert << "\t" << ins << std::endl;
      std::string mispredicted_target = convert.str();
      PushASCIIDescriptor(threadid, mispredicted_target);
      //reset iCnt, bCnt
      counters->ResetCounters();
   }
   PIN_ReleaseLock(&shared_lock);
}
VOID Shared_IndirectCall_Dis(const THREADID threadid, const ADDRINT addr, const ADDRINT target, const ADDRINT ret)
{
   //see if we can emit
   if(!CanEmit(threadid)) return;
   localCounters *counters = static_cast<localCounters*>(PIN_GetThreadData(tls_key, threadid));
   PIN_GetLock(&shared_lock, threadid+1);

   //increment bCnt and iCnt
   counters->IncrementiCnt();
   counters->IncrementbCnt();

   ADDRINT index = SharedTRaptor->GetIBTBindex(addr);
 
   //check if iBTB is hit or if not using iBTB
   BOOL correct = SharedTRaptor->IBTBisHit(addr, index, target);  
   //increment statistics for IndUB
   IncrementBranchStatistics(Target, correct);

   if ( !correct )
   {
      std::string ins;
      getAssemblyString(ins, threadid, addr);

      //write ASCII string
      std::ostringstream convert;
      convert << static_cast<UINT16>(counters->tid) << ", " << counters->bCnt;
      convert << ", T, 0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << target;
      convert << "\t" << ins << std::endl;
      std::string mispredicted_target = convert.str();
      PushASCIIDescriptor(threadid, mispredicted_target);
      //reset iCnt, bCnt
      counters->ResetCounters();
   }
   
   SharedTRaptor->PushRAS(ret);
   PIN_ReleaseLock(&shared_lock);
}
//other indirect branch
VOID Shared_OtherUnconditionalIndirect_Dis(const THREADID threadid, const ADDRINT addr, const ADDRINT target, const ADDRINT ret)
{
   //see if we can emit
   if(!CanEmit(threadid)) return;
   localCounters *counters = static_cast<localCounters*>(PIN_GetThreadData(tls_key, threadid));
   PIN_GetLock(&shared_lock, threadid+1);

   //increment bCnt and iCnt
   counters->IncrementiCnt();
   counters->IncrementbCnt();

   //get iBTB index 
   ADDRINT index = SharedTRaptor->GetIBTBindex(addr);

   //check if iBTB is hit or if not using iBTB
   BOOL correct = SharedTRaptor->IBTBisHit(addr, index, target);  
   //increment statistics for IndUB
   IncrementBranchStatistics(Target, correct);

   if ( !correct )
   {
      std::string ins;
      getAssemblyString(ins, threadid, addr);

      //write ASCII string
      std::ostringstream convert;
      convert << static_cast<UINT16>(counters->tid) << ", " << counters->bCnt;
      convert << ", T, 0x" << std::hex << std::setw(ADDRWID) << std::setfill('0') << target;
      convert << "\t" << ins << std::endl; 
      std::string mispredicted_target = convert.str();
      PushASCIIDescriptor(threadid, mispredicted_target);
      //reset iCnt, bCnt
      counters->ResetCounters();
   }
   PIN_ReleaseLock(&shared_lock);
}
/*------------------------*/
/* Bin Analysis routines */
/*------------------------*/
VOID Shared_DirectConditional_Bin(const THREADID threadid, const ADDRINT addr, const BOOL taken)
{
   //see if we can emit
   if(!CanEmit(threadid)) return;
   localCounters *counters = static_cast<localCounters*>(PIN_GetThreadData(tls_key, threadid));
   PIN_GetLock(&shared_lock, threadid+1);

   
   //increment bCnt and iCnt
   counters->IncrementiCnt();
   counters->IncrementbCnt();

   ADDRINT index = SharedTRaptor->GetGSHAREindex(addr);

   //see if prediction is correct
   BOOL correct = SharedTRaptor->OutcomePredictionIsCorrect(index, taken);

   //update statistics for DirCB
   IncrementBranchStatistics(Outcome, correct);

   //Emit descriptor if prediction is wrong or doesn't use GShare
   if( !correct )
   {
      BinaryDescriptorTableEntry mispredicted_outcome;
      mispredicted_outcome.tid = counters->tid;
      mispredicted_outcome.bCnt = counters->bCnt;
      mispredicted_outcome.type = Outcome;
      PushBinDescriptor(threadid, mispredicted_outcome);
      //reset iCnt, bCnt
      counters->ResetCounters();
   }
   //update TRaptor's GSHARE
   SharedTRaptor->UpdateGSHARE(index, taken);
   PIN_ReleaseLock(&shared_lock);
}
//return instruction, need to check RAS
VOID Shared_Ret_Bin(const THREADID threadid, const ADDRINT addr, const ADDRINT target)
{
   //see if we can emit
   if( !CanEmit(threadid) ) return;
   localCounters *counters = static_cast<localCounters*>(PIN_GetThreadData(tls_key, threadid));
   PIN_GetLock(&shared_lock, threadid+1);

   //increment bCnt and iCnt
   counters->IncrementiCnt();
   counters->IncrementbCnt();

   //check if RAS hits or doesn't use RAS
   BOOL correct = SharedTRaptor->RASHit(target);

   //increment statistics for IndUB
   IncrementBranchStatistics(Target, correct);

   if( !correct )
   {
      BinaryDescriptorTableEntry mispredicted_outcome;
      mispredicted_outcome.tid = counters->tid;
      mispredicted_outcome.bCnt = counters->bCnt;
      mispredicted_outcome.targetAddress = target;
      mispredicted_outcome.type = Target;
      PushBinDescriptor(threadid, mispredicted_outcome);
      //reset iCnt, bCnt
      counters->ResetCounters();
   }
   PIN_ReleaseLock(&shared_lock);
}
VOID Shared_IndirectCall_Bin(const THREADID threadid, const ADDRINT addr, const ADDRINT target, const ADDRINT ret)
{
   //see if we can emit
   if(!CanEmit(threadid)) return;
   localCounters *counters = static_cast<localCounters*>(PIN_GetThreadData(tls_key, threadid));
   PIN_GetLock(&shared_lock, threadid+1);

   //increment bCnt and iCnt
   counters->IncrementiCnt();
   counters->IncrementbCnt();

   ADDRINT index = SharedTRaptor->GetIBTBindex(addr);
 
   //check if iBTB is hit or if not using iBTB
   BOOL correct = SharedTRaptor->IBTBisHit(addr, index, target);  
   //increment statistics for IndUB
   IncrementBranchStatistics(Target, correct);

   if ( !correct )
   {
      BinaryDescriptorTableEntry mispredicted_outcome;
      mispredicted_outcome.tid = counters->tid;
      mispredicted_outcome.bCnt = counters->bCnt;
      mispredicted_outcome.targetAddress = target;
      mispredicted_outcome.type = Target;
      PushBinDescriptor(threadid, mispredicted_outcome);
      //reset iCnt, bCnt
      counters->ResetCounters();
   }
   
   SharedTRaptor->PushRAS(ret);
   PIN_ReleaseLock(&shared_lock);
}
//other indirect branch
VOID Shared_OtherUnconditionalIndirect_Bin(const THREADID threadid, const ADDRINT addr, const ADDRINT target, const ADDRINT ret)
{
   //see if we can emit
   if(!CanEmit(threadid)) return;
   localCounters *counters = static_cast<localCounters*>(PIN_GetThreadData(tls_key, threadid));
   PIN_GetLock(&shared_lock, threadid+1);

   //increment bCnt and iCnt
   counters->IncrementiCnt();
   counters->IncrementbCnt();

   //get iBTB index 
   ADDRINT index = SharedTRaptor->GetIBTBindex(addr);

   //check if iBTB is hit or if not using iBTB
   BOOL correct = SharedTRaptor->IBTBisHit(addr, index, target);  
   //increment statistics for IndUB
   IncrementBranchStatistics(Target, correct);

   if ( !correct )
   {
      BinaryDescriptorTableEntry mispredicted_outcome;
      mispredicted_outcome.tid = counters->tid;
      mispredicted_outcome.bCnt = counters->bCnt;
      mispredicted_outcome.targetAddress = target;
      mispredicted_outcome.type = Target;
      PushBinDescriptor(threadid, mispredicted_outcome);
      //reset iCnt, bCnt
      counters->ResetCounters();
   }
   PIN_ReleaseLock(&shared_lock);
}

/*--------------------------------------*/
/* Other Analaysis Calls */
/*--------------------------------------*/
VOID Shared_Sig(THREADID threadid, CONTEXT_CHANGE_REASON reason, const CONTEXT *ctxtFrom, CONTEXT *ctxTo, INT32 sig, VOID *v)
{
   if( !CanEmit(threadid) ) return;
   localCounters *counters = static_cast<localCounters*>(PIN_GetThreadData(tls_key, threadid));
   PIN_GetLock(&shared_lock, threadid+1);

   //From the Pin user guide
   //1. If the tool acquires any locks from within a Pin call-back, it must release those locks before returning from that call-back.
   //Holding a lock across Pin call-backs violates the hierarchy with respect to the Pin internal locks.
   //
   //2. If the tool calls a Pin API from within a Pin call-back or analysis routine, it should not hold any tool locks when calling the
   //API. Some of the Pin APIs use the internal Pin locks so holding a tool lock before invoking these APIs violates the hierarchy with
   //respect to the Pin internal locks.

   counters->IncrementiCnt();  
 
   PIN_GetLock(&cout_lock, threadid+1);
   std::cout << "mcfTRaptor: Thread " << static_cast<UINT16>(counters->tid) << " received signal " << sig << std::endl;
   if(reason == CONTEXT_CHANGE_REASON_FATALSIGNAL)
   {
      std::cout << "mcfTRaptor: Fatal signal, exiting immediately..." << std::endl;
      PIN_GetLock(&table_lock, threadid+1);
      closeOutput();
      PIN_ReleaseLock(&table_lock);
      PIN_ReleaseLock(&cout_lock);
      PIN_ReleaseLock(&shared_lock);
      return;
   }
   PIN_ReleaseLock(&cout_lock);

   //get context
   ADDRINT from = PIN_GetContextReg(ctxtFrom, REG_INST_PTR);
   ADDRINT to = PIN_GetContextReg(ctxTo, REG_INST_PTR);

   if( ASCII.Value() )
   {
      std::ostringstream convert;
      convert << static_cast<UINT16>(counters->tid);
      convert << ", 0";
      convert << ", " << counters->iCnt;
      convert << ", " << std::hex << std::setw(ADDRWID) << std::setfill('0') << to;
   
      if( AnnotateDisassembly.Value() )
      {
         std::string ins;
         getAssemblyString(ins, threadid, from);
 
         if(reason==CONTEXT_CHANGE_REASON_SIGNAL)
            ins = "\tSignal At: " + ins;
         else if(reason==CONTEXT_CHANGE_REASON_SIGRETURN)
            ins = "\tReturning from signal handler: " + ins;
         convert << ins;
      }
      convert << std::endl;  
      std::string except = convert.str();
      PushASCIIDescriptor(threadid, except);
   }
   else
   {
      BinaryDescriptorTableEntry binEntry;
      binEntry.type = Exception;
      binEntry.tid = counters->tid;
      binEntry.iCnt = counters->iCnt;
      binEntry.targetAddress = to;
      PushBinDescriptor(threadid, binEntry);
   } 
   counters->ResetCounters();
   PIN_ReleaseLock(&shared_lock);
}

//for direct calls we still need to push target on the RAS
VOID Shared_Direct_Call(THREADID threadid, ADDRINT addr, ADDRINT ret)
{
   if(!CanEmit(threadid)) return;
   localCounters *counters = static_cast<localCounters*>(PIN_GetThreadData(tls_key, threadid));
   PIN_GetLock(&shared_lock, threadid+1);
   counters->IncrementiCnt();
   SharedTRaptor->PushRAS(ret);
   PIN_ReleaseLock(&shared_lock);
}
//increment iCnt for all other other instructions
VOID Shared_iCnt_Increment(THREADID threadid)
{
   if(!CanEmit(threadid)) return;
   localCounters *counters = static_cast<localCounters*>(PIN_GetThreadData(tls_key, threadid));
   PIN_GetLock(&shared_lock, threadid+1);
   counters->IncrementiCnt();
   PIN_ReleaseLock(&shared_lock);
}

